from cbp.node import FactorNode

from .hmm_builder import HMMBuilder


class HMMSimBuilder(HMMBuilder):
    def __init__(self, length, simulator, policy,
                 random_seed=1):
        super().__init__(length, simulator.record.state_num, policy, random_seed)
        self.simulator = simulator
        self.cnt_constrained_node = 0

    def fix_initpotential(self, potential=None):
        first_node = self.graph.get_node(f"VarNode_{0:03d}")
        if potential is None:
            potential = self.simulator.get_hidden_margin(0)
        first_node.potential = potential

    def add_constrained_node(self, probability=None):
        prob = self.simulator.get_fix_margin(
            self.cnt_constrained_node)
        self.cnt_constrained_node += 1
        return super().add_constrained_node(prob)

    def add_factor(self, name_list, is_conv=False):
        """add factor to hmm graph

        :param name_list: connected node
        :type name_list: list
        :param is_conv: is emit or not , defaults to False-- transition
        :type is_conv: bool, optional
        :return: FactorNode
        :rtype: cbp.FactorNode
        """
        if is_conv:  # emit
            potential = self.simulator.get_emission_potential()
        else:
            potential = self.simulator.get_transition_potential()
        factornode = FactorNode(name_list, potential)
        self.graph.add_factornode(factornode)
        return factornode

    def add_branch(self, head_node=None, is_constrained=False,
                   prob=None, is_conv=False):
        if is_conv:  # emit
            return super().add_branch(head_node, True, prob, is_conv)
        return super().add_branch(head_node, False, prob, is_conv)

    def example(self, num_sample):
        self.simulator.reset()
        self.simulator.example(num_sample)



from collections import namedtuple
import numpy as np


class Message(namedtuple('Message', ['sender', 'val'])):
    def __new__(cls, sender, val):
        return super(Message, cls).__new__(cls, sender, val / np.sum(val))



import unittest

import numpy as np
from cbp.graph.coef_policy import bp_policy

from .utils import two_node_tree


class TestGraph(unittest.TestCase):
    def setUp(self):
        self.graph = two_node_tree()
        self.graph.coef_policy = bp_policy

    def test_marginal_brutal_force(self):
        self.graph.exact_marginal()
        node_equal = np.isclose(np.array([0.2, 0.8]),
                                self.graph.get_node("VarNode_000").bfmarginal)
        self.assertTrue(all(node_equal))
        node_equal = np.isclose(np.array([0.24, 0.76]),
                                self.graph.get_node("VarNode_001").bfmarginal)
        self.assertTrue(all(node_equal))

    def test_marginal_bp(self):
        self.graph.run_bp()
        node_equal = np.isclose(np.array([0.2, 0.8]),
                                self.graph.get_node("VarNode_000").marginal())
        self.assertTrue(all(node_equal))
        node_equal = np.isclose(np.array([0.24, 0.76]),
                                self.graph.get_node("VarNode_001").marginal())
        self.assertTrue(all(node_equal))

    def test_cnp_coef(self):
        self.graph.bake()
        self.graph.get_node("VarNode_000").node_coef = 1.0 / 4
        self.graph.get_node("VarNode_001").node_coef = 1.0 / 4
        self.graph.get_node("FactorNode_000").node_coef = 1.0 / 2
        self.graph.get_node("FactorNode_000").set_i_alpha(
            "VarNode_000", 1.0 / 4)
        self.graph.get_node("FactorNode_000").set_i_alpha(
            "VarNode_001", 1.0 / 4)
        self.graph.norm_product_bp()
        node_equal = np.isclose(np.array([0.2, 0.8]),
                                self.graph.get_node("VarNode_000").marginal(),
                                atol=1e-3, rtol=1e-2)
        self.assertTrue(all(node_equal))
        node_equal = np.isclose(np.array([0.24, 0.76]),
                                self.graph.get_node("VarNode_001").marginal(),
                                atol=1e-3, rtol=1e-2)
        self.assertTrue(all(node_equal))



import numpy as np
from numpy.random import RandomState
from scipy.ndimage import gaussian_filter
from cbp.utils.np_utils import batch_normal_angle

from .migr_visualizer import MigrVisualizer
from .hmm_simulator import HMMSimulator, PotentialType


class MigrSimulator(
        HMMSimulator):  # pylint: disable=too-many-instance-attributes
    def __init__(self, time_step, d_col, d_row, random_seed):  # pylint: disable=too-many-function-args
        super().__init__(time_step, d_col * d_row, d_col * d_row, random_seed)
        self.d_row = d_row
        self.d_col = d_col
        self.rng = RandomState(random_seed)

        self._sim = {
            "angle_wind": np.pi / 2,
            "sensor_sigma": 4,
            "num_sensor": self.record.state_num,
            "destination": (self.d_row, self.d_col)
        }
        self.visualizer = MigrVisualizer(self.d_row, self.d_col)

    def compile(self):
        """register various potential for simulation
        """
        self._produce_transition_potential()
        self._produce_sensor_potential()

        init_potential = np.zeros(self.record.state_num)
        init_potential[0] = 1
        self.register_potential(PotentialType.INIT, init_potential)

    def _transition_factor_goal(self, cur_col, cur_row):
        col, row = np.meshgrid(np.arange(self.d_col), np.arange(self.d_row))
        diff_row = row - cur_row
        diff_col = col - cur_col

        angle_goal = np.arctan2(
            self._sim["destination"][0] - 1 - cur_row,
            self._sim["destination"][1] - 1 - cur_col)

        angle_matrix = np.arctan2(diff_row, diff_col)

        goal_matrix = np.abs(
            batch_normal_angle(
                angle_matrix - angle_goal))
        return 5 * goal_matrix

    def _transition_factor_wind(self, cur_col, cur_row):
        col, row = np.meshgrid(np.arange(self.d_col), np.arange(self.d_row))
        diff_row = row - cur_row
        diff_col = col - cur_col

        angle_matrix = np.arctan2(diff_row, diff_col)

        wind_matrix = np.abs(
            batch_normal_angle(
                angle_matrix -
                self._sim["angle_wind"]))

        return 3 * wind_matrix

    def _transition_factor_dist(self, cur_col, cur_row):
        col, row = np.meshgrid(np.arange(self.d_col), np.arange(self.d_row))
        diff_row = row - cur_row
        diff_col = col - cur_col
        dist_matrix = np.sqrt(np.power(diff_row, 2)
                              + np.power(diff_col, 2))

        return 1.6 * dist_matrix

    def _transition_logistic_regression(self, cur_col, cur_row):
        dist_matrix = self._transition_factor_dist(cur_col, cur_row)
        wind_matrix = self._transition_factor_wind(cur_col, cur_row)
        goal_matrix = self._transition_factor_goal(cur_col, cur_row)

        exponent = -dist_matrix - wind_matrix - goal_matrix
        exponent[cur_row, cur_col] += 1

        cur_prob = np.exp(exponent)

        return cur_prob.flatten() / np.sum(cur_prob)

    def _produce_transition_potential(self):
        potential = []

        for cur_row in range(self.d_row):
            for cur_col in range(self.d_col):
                potential.append(
                    self._transition_logistic_regression(
                        cur_col, cur_row))
        self.register_potential(PotentialType.TRANSITION, np.array(
            potential).reshape(self.record.state_num, self.record.state_num))

    def _produce_sensor_potential(self):
        potential = []
        for i in range(self.d_row):
            for j in range(self.d_col):
                empty_matrix = np.zeros((self.d_row, self.d_col))
                empty_matrix[i, j] = 100.0
                result = gaussian_filter(
                    empty_matrix, sigma=self._sim["sensor_sigma"])
                potential.append(result.flatten() / np.sum(result))

        self.register_potential(PotentialType.EMISSION, np.array(
            potential).reshape(self.record.state_num, self.record.state_num))

    def viz_emission_potential(self):
        self.visualizer.potential_heatmap(
            self.record[PotentialType.EMISSION],
            title="sensor_potential",
            path=f"{self.path}/sensor_potential")

    def viz_trans_potential(self):
        self.visualizer.potential_heatmap(
            self.record[PotentialType.TRANSITION],
            title="transition_potential",
            path=f"{self.path}/transition_potential")

    def viz_gt(self):
        self.visualizer.migration(self.record["traj"],
                                  **{"title": "bird traj",
                                     "path": f"{self.path}/gt",
                                     "ylabel": True,
                                     "xlabel": 'Ground Truth'})

    def viz_sensor(self):
        self.visualizer.migration(self.record["sensor"],
                                  **{"title": "bird traj",
                                     "path": f"{self.path}/sensor"})

    def viz_estm(self, estimated_marginal):
        for i in range(self.record.time_step):
            bins = self.record["num_sample"] * \
                estimated_marginal[i, :]
            png_name = f"{self.path}/estimated_{i}.png"
            self.visualizer.visualize_map_bins(
                bins, fig_name=png_name, xlabel='Estimated')



from pathlib import Path

from .wifi_simulator import WifiSimulator
from .hmm_sim_builder import HMMSimBuilder


class WifiHMMBuilder(HMMSimBuilder):
    """builder with wifi-type simulator. Sparse observation. pass the
    wifi-simulator to the `cbp.builder.HMMSimBuilder`
    """

    def __init__(self, length, grid_d, policy,  # pylint: disable=too-many-arguments
                 rand_seed=1, num_sensor=16, time_step=60):
        sim_name = f'wifi-timestep_{time_step}-d_{grid_d}-rs_{rand_seed}-ns_{num_sensor}'
        sim_path = Path(f"data/sim/{sim_name}/sim.pkl")
        if sim_path.is_file():
            print(f"loading simulator from {sim_path}")
            simulator = WifiSimulator.load(sim_name)
        else:
            simulator = WifiSimulator(time_step, grid_d, grid_d, rand_seed)#Changed 60 to time_step
            simulator.name = sim_name
            simulator.random_sensor(num_sensor)
            simulator.compile()
            simulator.sample(10000)
            simulator.viz_sensor()
            simulator.viz_gt()
            print(f"saving simulator in {simulator.path}")
            simulator.save()

        super().__init__(length, simulator, policy, rand_seed)



import unittest
from pathlib import Path
from cbp import builder
from cbp.graph.coef_policy import bp_policy


class TestPlotGraph(unittest.TestCase):
    def test_plot_graph(self):
        Path('data').mkdir(exist_ok=True)
        builder_func = {
            "line": builder.LineBuilder,
            "star": builder.StarBuilder,
            "hmm": builder.HMMBuilder,
            "tree": builder.TreeBuilder
        }
        for name, worker in builder_func.items():
            graph = worker(4, 4, bp_policy, rand_seed=1)()
            graph.plot(f"data/{name}.png")



import time

import numpy as np


def engine_loop(  # pylint: disable=too-many-arguments
        engine_fun,
        max_iter=5000000,
        tolerance=1e-2,
        error_fun=None,
        meassure_fun=None,
        isoutput=False,
        silent=False):
    check_step = 1
    epsilons = [np.inf] * check_step
    start = time.time()
    timer_record = []
    step = 0
    cur_marginals = meassure_fun()

    while (step < max_iter) and any(
            tolerance < np.array(epsilons[-check_step:])):
        last_marginals = cur_marginals
        step += 1

        engine_fun()
        cur_marginals = meassure_fun()
        epsilons.append(error_fun(cur_marginals, last_marginals))

        timer_record.append(time.time() - start)
        if not silent:
            print(
                f"epsilon: {epsilons[-1]:5.4f} | step: {step:5d} {'-'*10}")
            print(cur_marginals)
            print(last_marginals)
            print(epsilons)
        if isoutput:
            print(f'step: {step:5d} | verobose output {epsilons[-1]}')

    return epsilons[check_step:], step, timer_record


def compare_marginals(mar_1, mar_2):
    assert np.setdiff1d(mar_1.keys(), mar_2.keys()).size == 0
    return sum([np.sum(np.absolute(mar_1[k] - mar_2[k])) for k in mar_1.keys()])


def diff_max_marginals(mar_1, mar_2):
    assert np.setdiff1d(mar_1.keys(), mar_2.keys()).size == 0
    return np.max([np.sum(np.absolute(mar_1[k] - mar_2[k]))
                   for k in mar_1.keys()])



import unittest

import numpy as np
from cbp.graph.coef_policy import bp_policy

from .utils import six_node_graph


class TestBpSxiVar(unittest.TestCase):
    def setUp(self):
        self.graph = six_node_graph()

    def test_marginal_bp(self):
        self.graph.coef_policy = bp_policy
        self.graph.run_bp()
        self.graph.exact_marginal()

        isequal_list = []

        for item in self.graph.varnode_recorder.values():
            marginal_equal = np.isclose(item.bfmarginal, item.marginal())
            isequal_list.append(all(marginal_equal))

        self.assertTrue(all(isequal_list))



from functools import partial

from cbp.utils import (compare_marginals, diff_max_marginals,
                       engine_loop)
from cbp.configs.base_config import baseconfig

from .base_graph import BaseGraph
from .coef_policy import bp_policy
from .graph_utils import itsbp_inner_loop, find_link


class GraphModel(BaseGraph):
    def __init__(self, silent=True, epsilon=1,
                 coef_policy=bp_policy, config=baseconfig):
        super().__init__(config=config, silent=silent,
                         epsilon=epsilon, coef_policy=coef_policy)
        self.itsbp_outer_cnt = 0

    def init_cnp_coef(self):
        for node in self.nodes:
            node.cal_cnp_coef()

    def run_cnp(self):
        self.bake()
        return self.norm_product_bp()

    def run_bp(self):
        if self.coef_policy != bp_policy:  # pylint: disable=comparison-with-callable
            self.coef_policy = bp_policy
        self.bake()
        return self.itsbp()

    def norm_product_bp(self, max_iter=5000000, tolerance=1e-5, error_fun=None):
        if error_fun is None:
            error_fun = diff_max_marginals
        self.init_cnp_coef()
        self.first_belief_propagation()
        return self.engine_loop(
            max_iter=max_iter,
            engine_fun=self.parallel_message,
            tolerance=tolerance,
            error_fun=error_fun,
            isoutput=False)

    def engine_loop(  # pylint: disable= too-many-arguments
            self,
            engine_fun,
            max_iter=5000000,
            tolerance=1e-2,
            error_fun=None,
            isoutput=False):
        if error_fun is None:
            error_fun = compare_marginals

        epsilons, step, _ = engine_loop(
            engine_fun=engine_fun,
            max_iter=max_iter,
            tolerance=tolerance,
            error_fun=error_fun,
            meassure_fun=self.export_convergence_marginals,
            isoutput=isoutput,
            silent=self.silent
        )

        return epsilons, step

    def itsbp(self):
        """run sinkhorn or iterative scaling inference

        :return: [description]
        :rtype: [type]
        """
        self.init_cnp_coef()
        self.first_belief_propagation()

        inner_bind = partial(self.parallel_message, False)
        self.engine_loop(inner_bind,
                         tolerance=1e-1,
                         error_fun=diff_max_marginals)

        return self.engine_loop(self.itsbp_outer_loop,
                                tolerance=1e-4,
                                error_fun=diff_max_marginals,
                                isoutput=False)

    def its_next_looplink(self):
        target_node = self.leaf_nodes[self.itsbp_outer_cnt]

        next_node = self.leaf_nodes[(
            self.itsbp_outer_cnt + 1) % len(self.leaf_nodes)]

        self.itsbp_outer_cnt = self.cfg.itsbp_schedule(
            self.itsbp_outer_cnt, self.leaf_nodes)
        return target_node, find_link(target_node, next_node)

    def itsbp_outer_loop(self):
        for _ in range(len(self.leaf_nodes)):
            _, loop_link = self.its_next_looplink()
            inner_fun = partial(itsbp_inner_loop, loop_link, self.silent)

            self.engine_loop(inner_fun,
                             tolerance=1e-3,
                             error_fun=diff_max_marginals,
                             isoutput=False)

    def parallel_message(self, run_constrained=True):
        for target_var in self.varnode_recorder.values():
            # sendind in messages from factors
            target_var.sendin_message(self.silent)

            if run_constrained or (not target_var.isconstrained):
                target_var.sendout_message(self.silent)

    def two_pass(self):
        # TODO: remove!
        self.init_cnp_coef()
        self.first_belief_propagation()
        for node in self.nodes:
            node.marked = False

        for node in self.nodes:
            if len(node.connections) == 1:
                root_node = node

        self.send_from(root_node)
        self.send_out(root_node)

    def send_from(self, node):
        node.marked = True
        for cur_node in node.connected_nodes.values():
            if not cur_node.marked:
                self.send_from(cur_node)
                cur_node.send_message(node)

    def send_out(self, node):
        node.marked = False
        for cur_node in node.connected_nodes.values():
            if cur_node.marked:
                node.send_message(cur_node)
                self.send_out(cur_node)



from .base_builder import BaseBuilder


class HMMBuilder(BaseBuilder):
    def __init__(self, length, node_dim, policy, rand_seed=1):
        self.hmm_length = length
        self.__step = 0
        super().__init__(node_dim, policy, rand_seed)

    def step(self, time_stamp=None):
        """hmm go forward a step, time + 1

        :param time_stamp: timer for the current add node
        :type time_stamp: int
        """
        if time_stamp is None:
            time_stamp = self.__step
        assert time_stamp > 0
        self.add_branch(head_node=f"VarNode_{2*time_stamp-2:03d}")
        self.add_branch(is_constrained=True, is_conv=True)
        self.__step += 1

    def init_graph(self):
        self.add_trivial_node()
        self.add_branch(is_constrained=True, is_conv=True)
        self.__step = 1

        for i in range(1, self.hmm_length):
            self.step(time_stamp=i)



import unittest
import timeit

import numpy as np
from cbp.builder import HMMBuilder, HMMZeroBuilder, LineBuilder
from cbp.graph.coef_policy import bp_policy
from cbp.configs import TestConfig

from test.utils import sinkhorn_bp_equal, two_node_tree


def next_links(graph):
    _, link = graph.its_next_looplink()
    nodenames = [node.name for node in link]
    return nodenames


class TestITSbp(unittest.TestCase):
    def test_itsbp_get_loop_single(self):
        graph = two_node_tree()
        graph.bake()
        _, link = graph.its_next_looplink()
        nodenames = [node.name for node in link]
        expected_names = [
            "VarNode_000",
            "FactorNode_000",
            "VarNode_001"
        ]
        self.assertEqual(nodenames, expected_names)

    zero2one = [
        f'VarNode_{1:03d}',
        f'FactorNode_{0:03d}',
        f'VarNode_{0:03d}',
        f'FactorNode_{1:03d}',
        f'VarNode_{2:03d}',
        f'FactorNode_{2:03d}',
        f'VarNode_{3:03d}']
    one2two = [
        f'VarNode_{3:03d}',
        f'FactorNode_{2:03d}',
        f'VarNode_{2:03d}',
        f'FactorNode_{3:03d}',
        f'VarNode_{4:03d}',
        f'FactorNode_{4:03d}',
        f'VarNode_{5:03d}']
    two2zero = [
        f'VarNode_{5:03d}',
        f'FactorNode_{4:03d}',
        f'VarNode_{4:03d}',
        f'FactorNode_{3:03d}',
        f'VarNode_{2:03d}',
        f'FactorNode_{1:03d}',
        f'VarNode_{0:03d}',
        f'FactorNode_{0:03d}',
        f'VarNode_{1:03d}']

    def test_get_loop_forward(self):
        graph = HMMBuilder(3, 2, bp_policy)()
        graph.bake()
        self.assertEqual(next_links(graph), TestITSbp.zero2one)
        self.assertEqual(next_links(graph), TestITSbp.one2two)
        self.assertEqual(next_links(graph), TestITSbp.two2zero)
        self.assertEqual(next_links(graph), TestITSbp.zero2one)

    def test_get_loop_backward(self):
        graph = HMMBuilder(3, 2, bp_policy)()
        graph.cfg = TestConfig()
        graph.bake()
        self.assertEqual(next_links(graph), TestITSbp.zero2one)
        self.assertEqual(next_links(graph), TestITSbp.one2two)
        self.assertEqual(list(reversed(next_links(graph))), TestITSbp.one2two)
        self.assertEqual(list(reversed(next_links(graph))), TestITSbp.zero2one)
        self.assertEqual(next_links(graph), TestITSbp.zero2one)

    def _profile_hmm_schedule(self, cfg=None):
        rng = np.random.RandomState(1)
        for _ in range(1):
            num_node = int(rng.randint(13, 15))
            node_dim = int(rng.randint(4, 8))
            graph = HMMBuilder(num_node, node_dim, bp_policy)()
            if cfg:
                graph.cfg = cfg
            graph.run_bp()

    @unittest.skip("Expensive test!")
    def test_cmp_schedule(self):
        print(timeit.timeit(
            lambda: self._profile_hmm_schedule(
                TestConfig()),
            number=1))
        print(timeit.timeit(lambda: self._profile_hmm_schedule(), number=1))

    def test_itsbp_line(self):
        rng = np.random.RandomState(1)
        for _ in range(10):
            num_node = int(rng.randint(3, 10))
            node_dim = int(rng.randint(2, 5))
            self.graph = LineBuilder(num_node, node_dim, bp_policy)()
            self.graph.run_bp()
            self.graph.sinkhorn()
            self.assertTrue(
                all(sinkhorn_bp_equal(self.graph, len_node=num_node)))

    def test_itsbp_hmm(self):
        rng = np.random.RandomState(1)
        for i in range(10):
            num_node = int(rng.randint(3, 6))
            node_dim = int(rng.randint(2, 5))
            self.graph = HMMBuilder(num_node, node_dim, bp_policy)()
            self.graph.sinkhorn()
            self.graph.run_bp()
            self.assertTrue(
                all(sinkhorn_bp_equal(self.graph, len_node=num_node)))

    def test_zero_hmm(self):
        graph = HMMZeroBuilder(3, 3, bp_policy)()
        graph.sinkhorn()
        graph.run_bp()
        self.assertTrue(all(sinkhorn_bp_equal(graph, len_node=3)))


if __name__ == '__main__':
    unittest.main()



from cbp.builder import WifiSimulator, WifiHMMBuilder
from cbp.graph.coef_policy import bp_policy
import unittest


class TestWifiSim(unittest.TestCase):
    def test_wifi_sim(self):
        sim = WifiSimulator(3, 5, 5, 3)
        sim.random_sensor(10)
        sim.compile()
        sim.sample(1000)
        sim.viz_gt()
        sim.viz_sensor()
        sim.viz_trans_potential()

    @unittest.skip("test build")
    def test_wifi_builder(self):
        builder = WifiHMMBuilder(
            length=20,
            grid_d=10,
            policy=bp_policy,
            rand_seed=5,
            num_sensor=64,
            time_step=20)
